#include "externals.hh"
#include "raymarcher.hh"
#include "profiling.hh"
#include "globals.hh"
#include "odbg.hh"


namespace {
	static char const* const Name_( "Raymarcher" );
}

#define PSTART() Globals::Profiler( ).start( Name_ )
#define PEND() do { glFinish( ); Globals::Profiler( ).end( Name_ ); } while ( 0 )


T_Raymarcher::T_Raymarcher(
		__rd__ const uint32_t width ,
		__rd__ const uint32_t height )
	: camera_( ) ,
		txOutput_( width , height , E_TexType::RGB16F ) ,
		txDepth_( width , height , E_TexType::R16F ) ,
		rtOutput_( T_RendertargetSetup( )
				.add( txOutput_ )
				.add( txDepth_ )
				.create( ) )
{
	Globals::ODbg( ).registerTexture( txOutput_ , E_ODbgMode::HDR , "Raymarched image" );
	Globals::ODbg( ).registerTexture( txDepth_ , E_ODbgMode::DEPTH , "Raymarched distance" );
	program_ = Globals::Shaders( ).pipeline({
			"fullscreen.v.glsl" , "scene.f.glsl" });
}


void T_Raymarcher::render( )
{
	if ( !rtOutput_.activate( ) || !program_.valid( ) ) {
		glClearColor( 0 , 0 , 0 , 1 );
		glClear( GL_COLOR_BUFFER_BIT );
		return;
	}

	PSTART( );
	program_.enable( );

	glClearColor( 0 , 0 , 0 , 1 );
	glClear( GL_COLOR_BUFFER_BIT );
	enum {
		U_TIME			= 0 ,
		U_RESOLUTION		= 1 ,
		U_CAM_POS		= 2 ,
		U_LOOK_AT		= 3 ,
		U_CAM_UP		= 4 ,
		U_NEAR_PLANE		= 5 ,
		U_RAYMARCHER		= 6 ,
		U_FOG			= 7 ,
		U_CORRECTION		= 8
	};

	const auto id( program_.program( E_ShaderType::FRAGMENT ) );
	glProgramUniform1f( id , U_TIME , 0 );
	glProgramUniform2f( id , U_RESOLUTION , rtOutput_.width( ) , rtOutput_.height( ) );

	glProgramUniform3fv( id , U_CAM_POS , 1 , &camera_.pos.x );
	glProgramUniform3fv( id , U_LOOK_AT , 1 , &camera_.lookAt.x );
	glProgramUniform3fv( id , U_CAM_UP , 1 , &camera_.up.x );
	glProgramUniform1f( id , U_NEAR_PLANE , camera_.np );

	glProgramUniform4f( id , U_RAYMARCHER , rmIterations , rmStep ,
			rmEpsilon , rmMaxDist );
	glProgramUniform1f( id , U_FOG , fog );
	glProgramUniform1i( id , U_CORRECTION , correction_ );

	glDrawArrays( GL_TRIANGLE_STRIP , 0 , 4 );

	PEND( );
}

void T_Raymarcher::makeUI( )
{
	camera_.makeUI( );

	if ( ImGui::CollapsingHeader( "Raymarcher" ) ) {
		ImGui::DragInt( "Iterations" , &rmIterations , .01 , 1 , 512 );
		ImGui::DragFloat( "Step" , &rmStep , .00005 , .1 , 2 );
		ImGui::DragFloat( "Max. dist." , &rmMaxDist , .1f ,
				0.01 , 1000.0 , "%.2f" );
		if ( ImGui::DragFloat( "Epsilon" , &epsLog , .01f ,
				-10 , -0.5 , "10 ^ %.2f" ) ) {
			rmEpsilon = pow( 10 , epsLog );
		}
		ImGui::DragInt( "Correction" , &correction_ , .001 , 0 , 50 );
		ImGui::DragFloat( "Fog" , &fog , .000001 , 0 , 1 , "%.5f" );
	}
}
